/*
 *  polyMath.cpp
 *  fp
 *
 *  Created by Robert Delaney on 5/17/13.
 *  Copyright 2013 Bob Delaney's Science Software. All rights reserved.
 *
 */

#include "matq.h"
#include "matqMath.h"
#include "poly.h"
#include "polyMath.h"
#include "polyConv.h"

// remove zeroes continuos from high order down
poly polyNormalize(const poly& x)
{
	poly		z;
	
	z = x;
	while(!z.array[z.deg] && z.deg>0)
		z.deg--;
	
	return z;
	
}/* polyNormalize */


// remove low order zeroes, effectively factoring out powers of x in poly(x)
poly polyReduce(const poly& x)
{
	INT32		i;
	poly		z;
	
	z = x;
	
	while(!z.array[0].num.n && z.deg>0)
	{
		for(i=0;i<z.deg;i++)
			z.array[i] = z.array[i+1];
		
		z.deg--;
	}
	
	return z;
	
}/* polyReduce */


// converts to integer coefficients
void polyOverZ(poly& z, const poly& x)
{
	mb			c;
	INT32		i;	
	poly		zt;
	
	zt = x;
	
	// get products of all denometers
	c = 1;
	
	for(i=0;i<=zt.deg;i++)
		c = c*zt.array[i].den;
	
	// convert to integers
	for(i=0;i<=zt.deg;i++)
		zt.array[i] = c*zt.array[i];
	
	// factor out largest common integer
	if(!zt.deg && !zt.array[0])
	{
		z = zt;
		return;
	}
	
	if(!zt.deg)
	{
		zt.array[0] = 1;
		z = zt;
		return;
	}
	
	if(zt.deg==1)
	{
		c = gcd(zt.array[0].num, zt.array[1].num);
		if(zt.array[1]>0)
		{
			zt.array[0] = zt.array[0] / c;
			zt.array[1] = zt.array[1] / c;
		}
		else
		{
			zt.array[0] = -zt.array[0] / c;
			zt.array[1] = -zt.array[1] / c;
		}
		z = zt;
		return;
	}
	
	// now zt.deg>1
	c = gcd(zt.array[0].num, zt.array[1].num);
	
	if(c==1)
	{
		z = zt;
		return;
	}
	
	for(i=2;i<=zt.deg;i++)
	{
		c = gcd(c, zt.array[i].num);
		if(c==1)
			break;
	}
	
	if(c>1)
	{
		if(zt.array[zt.deg]>0)
		{
			for(i=0;i<=zt.deg;i++)
				zt.array[i] = zt.array[i] / c;
		}
		else
		{
			for(i=0;i<=zt.deg;i++)
				zt.array[i] = -zt.array[i] / c;
		}
	}
		
	z = zt;
	return;
	   
}/* polyOverZ */


poly polyOverZ(const poly& x)
{
	poly		z;
	
	polyOverZ(z, x);
	
	return z;
	
}/* polyOverZ */


// takes derivative of poly x
poly polyDeriv(const poly& x)
{
	poly		z;
	INT32		i;
	
	if(!x.deg)
	{
		z = 0;
		return z;
	}
	
	z.deg = x.deg - 1;
	z.array = (bf*)malloc((z.deg+1)*sizeof(bf));
	
	for(i=1;i<=x.deg;i++)
		z.array[i-1] = i*x.array[i];
	
	z = polyNormalize(z);
	
	return z;
	
}/* polyDeriv */


// returns 1 if x=y, 0 if x!=y
int compare(const poly& x, const poly& y)
{
	poly		z;
	
	z = x - y;
	
	if(!z.deg && !z.array[0])
		return 1;
	
	return 0;
	
}/* compare */


void add(poly& z, const poly& x, const poly& y)
{
	INT32		i;
	poly		zt;
	
	zt.deg = max(x.deg, y.deg);
	
	zt.array = (bf*)malloc((zt.deg+1)*sizeof(bf));
	
	if(x.deg>=y.deg)
	{
		for(i=0;i<=y.deg;i++)
			zt.array[i] = x.array[i] + y.array[i];
		
		for(i=y.deg+1;i<=x.deg;i++)
			zt.array[i] = x.array[i];
	}
	else
	{
		for(i=0;i<=x.deg;i++)
			zt.array[i] = x.array[i] + y.array[i];
		
		for(i=x.deg+1;i<=y.deg;i++)
			zt.array[i] = y.array[i];
	}
	
	zt = polyNormalize(zt);
	
	z = zt;

}/* add */


void sub(poly& z, const poly& x, const poly& y)
{
	INT32		i;
	poly		zt;
	
	zt.deg = max(x.deg, y.deg);
	
	zt.array = (bf*)malloc((zt.deg+1)*sizeof(bf));
	
	if(x.deg>=y.deg)
	{
		for(i=0;i<=y.deg;i++)
			zt.array[i] = x.array[i] - y.array[i];
		
		for(i=y.deg+1;i<=x.deg;i++)
			zt.array[i] = x.array[i];
	}
	else
	{
		for(i=0;i<=x.deg;i++)
			zt.array[i] = x.array[i] - y.array[i];
		
		for(i=x.deg+1;i<=y.deg;i++)
			zt.array[i] = -y.array[i];
	}
	
	zt = polyNormalize(zt);
	
	z = zt;
	
}/* sub */


void mul(poly& z, const poly& x, const poly& y)
{
	INT32		i, j;
	poly		zt;
	
	zt.deg = x.deg + y.deg;
	
	zt.array = (bf*)malloc((zt.deg+1)*sizeof(bf));
	for(i=0;i<=zt.deg;i++)
		zt.array[i] = 0;
	
	for(i=0;i<=x.deg;i++)
		for(j=0;j<=y.deg;j++)
			zt.array[i+j] = zt.array[i+j] + x.array[i]*y.array[j];
	
	zt = polyNormalize(zt);
	
	z = zt;

}/* mul */


// want q = x/y with remainder r so x = q*y + r
void divEuclid(poly& q, poly& r, const poly& x, const poly& y)
{
	poly		xt, yt;
	poly		qt, rt;
	bf			s, c;
	INT32		i;
	
	// the degree of qt is x.deg - y.deg 
	// the max degree of rt is x.deg
	
	qt.deg = x.deg - y.deg;
	
	if(qt.deg>=0)
	{
		qt.array = (bf*)malloc((qt.deg+1)*sizeof(bf));
		qt.array[0] = 1;
		for(i=1;i<=qt.deg;i++)
			qt.array[i] = 0;
	}
	else
	{
		q = 0;
		r = x;
		return;
	}
	
	if(!y.deg)
	{
		r = 0;
		if(!y.array[0])
		{
			qt.deg = 0;
			qt.array[0] = 0;
			q = qt;
			return;
		}
		
		for(i=0;i<=x.deg;i++)
			qt.array[i] = x.array[i] / y.array[0];
		q = qt;
		return;
	}
	
	xt = x;
	yt = y;
	xt = polyNormalize(xt);
	yt = polyNormalize(yt);
	
	rt = xt;
	c = yt.array[yt.deg];  // k(y)
	
	while(rt.deg>=yt.deg)
	{
		s = rt.array[rt.deg] / c;
		qt.array[rt.deg-yt.deg] = s;
		for(i=rt.deg;i>=rt.deg-yt.deg;i--)
			rt.array[i] = rt.array[i] - s*yt.array[i-rt.deg+yt.deg];
			
		rt.deg--;
		rt = polyNormalize(rt);
	}
	
	q = qt;
	q = polyNormalize(q);
	r = rt;
	
}/* divEuclid */


// want q = x/y with remainder r so x = q*y + r mod p
// assume x, y belong to GF(p)[x]
// we assume if highest degree coefficient of y is n that gcd(n, p) = 1
void divEuclid(poly& q, poly& r, const poly& x, const poly& y, double p)
{
	poly		xt, yt;
	poly		qt, rt;
	bf			s, c;
	INT32		i;
	
	// the degree of qt is x.deg - y.deg 
	// the max degree of rt is x.deg
	
	qt.deg = x.deg - y.deg;
	
	if(qt.deg>=0)
	{
		qt.array = (bf*)malloc((qt.deg+1)*sizeof(bf));
		qt.array[0] = 1;
		for(i=1;i<=qt.deg;i++)
			qt.array[i] = 0;
	}
	else
	{
		q = 0;
		r = x;
		return;
	}
	
	if(!y.deg)
	{
		r = 0;
		if(!y.array[0])
		{
			qt.deg = 0;
			qt.array[0] = 0;
			q = qt;
			return;
		}
		
		for(i=0;i<=x.deg;i++)
			qt.array[i] = x.array[i] / y.array[0];
		q = qt;
		return;
	}
	
	xt = x;
	yt = y;
	xt = polyNormalize(xt);
	yt = polyNormalize(yt);
	
	rt = xt;
	c = yt.array[yt.deg];  // k(y)
	invMod(c.num, c.num, p);
	
	while(rt.deg>=yt.deg)
	{
		//s = r.array[r.deg] / c;
		s = c * rt.array[rt.deg];
		s.num = s.num % p;
		qt.array[rt.deg-yt.deg] = s;
		for(i=rt.deg;i>=rt.deg-yt.deg;i--)
		{
			rt.array[i] = rt.array[i] - s*yt.array[i-rt.deg+yt.deg];
			rt.array[i].num = rt.array[i].num % p;
		}
		
		rt.deg--;
		rt = polyNormalize(rt);
	}
	
	q = qt;
	q = polyNormalize(q);
	r = rt;
	
}/* divEuclid */


void div(poly& z, const poly& x, const poly& y)
{
	poly		q, r;
	
	divEuclid(q, r, x, y);
	
	z = q;

}/* div */


void add(poly& z, const poly& x, const bf& y)
{
	poly		zt;
	
	zt = x;
	zt.array[0] = zt.array[0] + y;
	
	z = zt;
}


void sub(poly& z, const poly& x, const bf& y)
{
	poly		zt;
	
	zt = x;
	zt.array[0] = zt.array[0] - y;
	
	z = zt;
}


void mul(poly& z, const poly& x, const bf& y)
{
	INT32		i;
	poly		zt;
	
	zt = x;
	for(i=0;i<=zt.deg;i++)
		zt.array[i] = y*zt.array[i];
	
	zt = polyNormalize(zt);  // for y=0
	
	z = zt;
}


void div(poly& z, const poly& x, const bf& y)
{
	INT32		i;
	poly		zt;
	
	if(!y)
	{
		z = 0;
		return;
	}
	
	zt = x;
	for(i=0;i<=zt.deg;i++)
		zt.array[i] = zt.array[i] / y;
	
	z = zt;
}


void add(poly& z, const bf& x, const poly& y)
{
	add(z, y, x);
}


void sub(poly& z, const bf& x, const poly& y)
{
	sub(z, y, x);
	z = -z;
}


void mul(poly& z, const bf& x, const poly& y)
{
	mul(z, y, x);
}


void div(poly& z, const bf& x, const poly& y)
{
	z.deg = 0;
	z.array = (bf*)malloc((z.deg+1)*sizeof(bf));
	if(y.deg>0)
	{
		z.array[0] = 0;
		return;
	}
	
	// now y.deg = 0
	if(!y.array[0])
	{
		z.array[0] = 0;
		return;
	}
	
	z.array[0] = x / y.array[0];
}


poly gcd(const poly& x, const poly& y)
{
	poly		q, r, xt, yt;
	bool		hasFactorx;
	
	xt = polyReduce(x);
	yt = polyReduce(y);
	hasFactorx = false;
	if(xt.deg<x.deg && yt.deg<y.deg)
		hasFactorx = true;
	
	if(yt.deg>xt.deg)
	{
		r = xt;
		xt = yt;
		yt = r;
	}
		
	while(!(!yt.deg && !yt.array[0]))
	{
		divEuclid(q, r, xt, yt);  // xt = q*yt + r
		xt = yt;
		yt = r;
	}
	
	xt = polyOverZ(xt);
	
	if(xt.deg==0 && hasFactorx)
	{
		free(xt.array);
		xt.deg = 1;
		xt.array = (bf*)malloc((xt.deg+1)*sizeof(bf));
		xt.array[0] = 0;
		xt.array[1] = 1;
	}
	
	return xt;
	
}/* gcd */

// it is assumed that x, y already belong to GF(q)[x]
// and that if n = highest coefficient of effective dividing polynomial in divEuclid then gcd(n, q) = 1
poly gcd(const poly& x, const poly& y, double q)
{
	poly		p, r, xt, yt;
	bool		hasFactorx;
	
	xt = polyReduce(x);
	yt = polyReduce(y);
	hasFactorx = false;
	if(xt.deg<x.deg && yt.deg<y.deg)
		hasFactorx = true;
	
	if(yt.deg>xt.deg)
	{
		r = xt;
		xt = yt;
		yt = r;
	}
	
	while(!(!yt.deg && !yt.array[0]))
	{
		divEuclid(p, r, xt, yt, q);  // xt = q*yt + r mod q
		xt = yt;
		yt = r;
	}
	
	xt = polyOverZ(xt);
	
	if(xt.deg==0 && hasFactorx)
	{
		free(xt.array);
		xt.deg = 1;
		xt.array = (bf*)malloc((xt.deg+1)*sizeof(bf));
		xt.array[0] = 0;
		xt.array[1] = 1;
	}
	
	return xt;
	
}/* gcd */


// with x = q*y + r, r = x mod y
poly polyMod(const poly& x, const poly& y)
{
	poly		q, r;
	
	divEuclid(q, r, x, y);
	
	//r = polyOverZ(r);
	
	return r;
	
}/* polyMod */


// with x = q*y + r, r = x mod y mod q
poly polyMod(const poly& x, const poly& y, double q)
{
	poly		p, r, z;
	
	divEuclid(p, r, x, y);
	
	//r = polyOverZ(r);
	
	z = polyMod(r, q);
	
	z = polyNormalize(z); // mod might make high order coefficient zero
	
	return z;
	
}/* polyMod */


// assumes integer coefficients
// reduces each coefficient of x by taking mod q
poly polyMod(const poly& x, const mb& q)
{
	poly		xt;
	INT32		i;
	
	xt = x;
	
	for(i=0;i<=xt.deg;i++)
		xt.array[i].num = (xt.array[i].num % q);
	
	return xt;
	
}/* polyMod */

// assumes integer coefficients
poly polyMod(const poly& x, double q)
{
	poly		xt;
	INT32		i;
	
	xt = x;
	
	for(i=0;i<=xt.deg;i++)
		xt.array[i].num = (xt.array[i].num % q);
	
	return xt;
	
}/* polyMod */


// Yun's algorithm
// http://en.wikipedia.org/wiki/Square-free_polynomial
void polySquareFree(poly*& z, long& numPolys, const poly& x)
{
	poly		a, b, c, d;
	
	numPolys = 1;
	// x.deg is max number of factors
	z = (poly*)malloc(x.deg*sizeof(poly));
	if(x.deg<2)
	{
		z[0] =  x;
		return;
	}
	
	a = gcd(x, polyDeriv(x)); // a0
	z[numPolys-1] = a;
	b = x / a;
	c = polyDeriv(x) / a;
	d = c - polyDeriv(b);
	while(!(!b.deg && b.array[0]==1))
	{
		a = gcd(b, d); // a1, a2, ...
		if(a!=1)
		{
			numPolys++;
			z[numPolys-1] = a;
		}
		b = b / a;
		c = d / a;
		d = c - polyDeriv(b);
	}
	
}/* polySquareFree */


void findPowersOfSquareFreeFactors(char*& zString, long& numPolys, const poly& x)
{
	poly		*z;
	poly		zt, r, q, temp;
	long		i;
	long		*theExp;
	bool		isGood;
	long		length;
	char		*xString, *numString;
	
	polySquareFree(z, numPolys, x);
	
	if(numPolys<1)
		return;
	
	if(numPolys==1)
	{
		polyConvToString(zString, z[0]);
		return;
	}
	
	numString = (char*)malloc(10*sizeof(char));
	
	theExp = (long*)malloc(numPolys*sizeof(long));
	for(i=0;i<numPolys;i++)
		theExp[i] = 0;
	
	
	//zt contains powers of some or all of z[1], z[2], ...
	for(i=1;i<numPolys;i++)
	{
		zt = z[0];
		isGood = true;
		while(isGood && z[i].deg)
		{
			divEuclid(q, r, zt, z[i]);
			if(!r.deg && !r.array[0])
			{
				theExp[i]++;
				zt = q;
			}
			else
			{
				isGood = false;
			}
		}
	}
	
	length = 0;
	for(i=1;i<numPolys;i++)
	{
		polyConvToString(xString, z[i]);
		length = length + strlen(xString);
		free(xString);
	}
	length+=(8*numPolys);
	
	zString = (char*)malloc(length*sizeof(char));
	
	zString[0] = 0;
	
	for(i=1;i<numPolys;i++)
	{
		polyConvToString(xString, z[i]);
		strcat(zString, "(");
		strcat(zString, xString);
		strcat(zString, ")");
		if(theExp[i]>0)
		{
			strcat(zString, "^");
			sprintf(numString,"%ld", theExp[i]+1);
			strcat(zString, numString);
		}
		if(i!=(numPolys-1))
			strcat(zString, " ");
	}
	
	numPolys--;
	
	free(theExp);
	free(numString);
	
}/* findPowersOfSquareFreeFactors */


void findPowersOfSquareFreeFactors(poly*& z1, long*& theExp, long& numPolys, const poly& x)
{
	poly		*z;
	poly		zt, r, q, temp;
	long		i;
	bool		isGood;
	
	polySquareFree(z, numPolys, x);
	
	if(numPolys<=1)
	{
		z1 = (poly*)malloc(sizeof(poly));
		theExp = (long*)malloc(sizeof(long));
		z1[0] = x;
		theExp[0] = 1;
		return;
	}
	
	theExp = (long*)malloc((numPolys-1)*sizeof(long));
	for(i=0;i<(numPolys-1);i++)
		theExp[i] = 0;
	
	for(i=1;i<numPolys;i++)
	{
		zt = z[0]; //zt contains powers of some or all of z[1], z[2], ...
		isGood = true;
		while(isGood && z[i].deg)
		{
			divEuclid(q, r, zt, z[i]);
			if(!r.deg && !r.array[0])
			{
				theExp[i-1]++;
				zt = q;
			}
			else
			{
				isGood = false;
			}
		}
	}
	
	z1 = (poly*)malloc((numPolys-1)*sizeof(poly));
	
	for(i=0;i<(numPolys-1);i++)
	{
		z1[i] = z[i+1];
		theExp[i]++;
	}
	
	numPolys--;
	
	//free(z);
	
}/* findPowersOfSquareFreeFactors */


// z = x^p
void polyPow(poly& z, const poly& x, INT32 p)
{
	INT32				i, n;
	static UINT32		mask; // 0x80000000
	poly				xt, zt;
	static bool			initGood=false;
	
	if(!initGood)
	{
		mask = 1;
		mask = (mask<<31); // 0x80000000
		initGood = true;
	}
	
	zt = x;
	
	if(p<0)
	{
		zt.deg = 0;
		zt.array[0] = 0;
		z = zt;
		return;
	}
	
	if(!p)
	{
		zt.deg = 0;
		zt.array[0] = 1;
		z = zt;
		return;
	}
	
	if(!zt.deg)
	{
		zt.array[0] = power(zt.array[0].num, p);
		z = zt;
		return;
	}
	
	// now p>0
	xt = polyOverZ(x);
	zt = 1;
	n = NumBits(p);
	
	p = (p<<(32-n));  // go to leading 1 bit
	
	for(i=0;i<n;i++)
	{
		mul(zt, zt, zt); // zt^2
		if(p & mask)
			mul(zt, zt, xt);
		p = (p<<1);
	}
	
	z = zt;
	
}/* polyPow */


poly  polyPow(const poly& x, INT32 p)
{
	poly		z;
	
	polyPow(z, x, p);
	
	return z;
	
}/* polyPow */


// z = x^p mod y
void polyPowerMod(poly& z, const poly& x, INT32 p, const poly& y)
{
	INT32				i, n;
	static UINT32		mask; // 0x80000000
	poly				xt, zt;
	static bool			initGood=false;
	
	if(!initGood)
	{
		mask = 1;
		mask = (mask<<31); // 0x80000000
		initGood = true;
	}
	
	if(p<0)
	{
		z = 0;
		return;
	}
	
	if(!p)
	{
		zt = 1;
		z = polyMod(zt, y);
		return;
	}
	
	zt = x;
	
	if(!zt.deg)
	{
		zt.array[0] = power(zt.array[0].num, p);
		z = polyMod(zt, y);
		return;
	}
	
	// now p>0
	xt = polyOverZ(x);
	zt = 1;
	n = NumBits(p);
	
	p = (p<<(32-n));  // go to leading 1 bit
	
	for(i=0;i<n;i++)
	{
		mul(zt, zt, zt); // zt^2
		zt = polyMod(zt, y);
		if(p & mask)
		{
			mul(zt, zt, xt);
			zt = polyMod(zt, y);
		}
		p = (p<<1);
	}
	
	z = zt;
	
}/* polyPowerMod */


// z = x^p mod y mod q
void polyPowerMod(poly& z, const poly& x, INT32 p, const poly& y, double q)
{
	INT32				i, n;
	static UINT32		mask; // 0x80000000
	poly				xt, zt;
	static bool			initGood=false;
	
	if(!initGood)
	{
		mask = 1;
		mask = (mask<<31); // 0x80000000
		initGood = true;
	}
	
	if(p<0)
	{
		z = 0;
		return;
	}
	
	if(!p)
	{
		zt = 1;
		zt = polyMod(zt, y);
		z = polyMod(zt, q);
		return;
	}
	
	zt = x;
	
	if(!zt.deg)
	{
		zt.array[0] = power(zt.array[0].num, p);
		zt = polyMod(zt, y);
		z = polyMod(zt, q);
		return;
	}
	
	// now p>0
	xt = polyOverZ(x);
	zt = 1;
	n = NumBits(p);
	
	p = (p<<(32-n));  // go to leading 1 bit
	
	for(i=0;i<n;i++)
	{
		mul(zt, zt, zt); // zt^2
		zt = polyMod(zt, y);
		zt = polyMod(zt, q);
		if(p & mask)
		{
			mul(zt, zt, xt);
			zt = polyMod(zt, y);
			zt = polyMod(zt, q);
		}
		p = (p<<1);
	}
	
	z = zt;
	
}/* polyPowerMod */


poly polyPowerMod(const poly& x, INT32 p, const poly& y)
{
	poly		z;
	
	polyPowerMod(z, x, p, y);
	
	return z;
	
}/* polyPowerMod */


poly polyPowerMod(const poly& x, INT32 p, const poly& y, double q)
{
	poly		z;
	
	polyPowerMod(z, x, p, y, q);
	
	return z;
	
}/* polyPowerMod */

